;ISHIDO - Way of the Stones
;copyright by Bodo Maass <bodo.maass@privat.post.de>
;http://www.pagerealm.com/bodo
;
;this code is released under the terms of the GNU General Public License
;http://www.gnu.org/copyleft/gpl.html

#include "usgard.h"

.org 0
.db "Ishido 1.0",0

POUCH	= GRAPH_MEM	;reserve 72*2 = 144 Byte for pieces
BOARD =   GRAPH_MEM+146 ;reserve 96*2 = 192 Byte for Board
OTHERVAR = GRAPH_MEM + 338
Score	   = OTHERVAR + 0	;2 byte for score
Num4Ways   = OTHERVAR + 2	;2 byte for number of fourways
PiecesLeft = OTHERVAR + 4	;2 byte for number of pieces left on board
EndBonus   = OTHERVAR + 6	;2 byte for endbonus
;DispLegal  = OTHERVAR + 8	;1 byte flag indicates whether legal-icon is displayed
;PossCount  = OTHERVAR + 9	;1 byte for number of possible moves

BASE_SET = TEXT_MEM
BASE_FEATURES = TEXT_MEM+144

SrcAddr    = TEXT_MEM2		;temporary storage for addresses
DstAddr    = TEXT_MEM2 + 2
CursorPos  = TEXT_MEM2 + 4	;2 byte for cursor position
OldCPos    = TEXT_MEM2 + 6	;2 byte for old position of cursor as screenaddr
OldCPosY   = TEXT_MEM2 + 8      ;1 byte y value of old cursor position
NextPiecePtr  = TEXT_MEM2 + 9	;2 byte copy of current piece in pouch
PhaseNr    = TEXT_MEM2 +11	;1 byte for phase number
PhasePtr   = TEXT_MEM2 +12	;2 byte pointer to current cursor data
AnimCount  = TEXT_MEM2 +14	;1 byte counter for animation delay
AnimCount1 = TEXT_MEM2 +15	;1 byte for second animation counter
NBcount    = TEXT_MEM2 +16	;1 byte counter of occupied neighbours of current field
MatchFlags = TEXT_MEM2 +17	;1 byte flags to indicate matching cols/shapes
seed1	   = TEXT_MEM2 +18	;2 byte for seed1
seed2      = TEXT_MEM2 +20	;2 byte for seed2
NextFieldPtr = TEXT_MEM2 + 22	;2 byte pointer to next boardfield

;BoardAddr  = Text_MEM2 +23	;2 bytes for BoardAddr of current field


;***********************************************************************************************
;***************                INITIALIZATION ROUTINES                *************************
;***********************************************************************************************
Init_Pouch:
	ld hl, &oldSeeds
	ld de, seed1
	ldi		;load stored seed into working mem
	ldi
	ldi
	ldi


;create temporary ordered list of pieces (144 Byte)
	ld hl, BASE_SET
	ld bc, $200		;do the whole thing two times (b) and set c to zero
	ld a, 6		;for comparison

CreateOrderSet2:
	ld d, c		;start with col-0 & shape-0 (c should still be zero)
CreateOrderSet1:
	ld e, c
CreateOrderSet0:
	ld (hl), d	;store colour
	inc hl
	ld (hl), e	;store shape
	inc hl
	inc e
	cp e		;e==6 ?
	jr nz, CreateOrderSet0
	inc d
	cp d		;d==6 ?
	jr nz, CreateOrderSet1
	
	djnz CreateOrderSet2	;repeat once for two full sets of 36 pieces


; hl now points to TEXT_MEM + 144 and a still contains 6
	push hl
	pop ix		;ld ix, hl
	;ld b, a		;save the value of a
ColShapeInitLoop:	;create list of values to draw the first six pieces
	dec a
	ld (ix), a	;list of six ordered colours
	ld (ix+6), a	;list of six ordered shapes
	inc ix
	jr nz, ColShapeInitLoop		;the z flag should remain unchanged since dec a
	
	
;now we must draw the first six pieces such that each color and each shape
;is represented once
; hl still points to TEXT_MEM + 144 = BASE_FEATURES
	;ld hl, BASE_FEATURES
	ld de, POUCH-2
	call &ScrambleShortLst	;scramble first six colours
	
	ld hl, BASE_FEATURES+6
	ld de, POUCH-1
	call &ScrambleShortLst	;scramble first six shapes
;now we must take out the first six pieces from the baseset
	ld b, 6		;loopcounter
	ld hl, POUCH
	ld d, 0		;set d to zero for addition of hl and de
TakeFirstOutLoop:
	
	ld e, (hl)	;get colour
	inc hl
	ld a, e		;a=1*col
	add a, a	;2*col
	add a, a	;4*col
	add a, e	;5*col
	add a, e	;6*col
	ld e, (hl)	;get shape
	inc hl
	add a, e	;6*col + shape
	add a, a	;2*(6*col+shape) (for 16-bit values)
	
	push hl		;save pouch-pointer
	ld hl, BASE_SET
	
	ld e, a		;de now contains only the accu (since d=0)
	add hl, de	;hl=BASE_SET+2*(6*col+shape)
			
	ld (hl), 5	;set colour to black and keep the shape
	pop hl		;restore pouch-pointer
	djnz TakeFirstOutLoop
	
;now there are 72-6=66 pieces left in the base-set, which have to be scrambled
;and then copied to the pouch, starting from POUCH+12 (6 Pieces @ 2 bytes)
; hl now points to POUCH+12 = first element in dest-list
; bc is zero
; de points to BASE_SET + 137 = last element in source-list
	ld b, 66
ScrambleLongLoop:
	ld e, b		;loopcounter = upper limit for rand
	push bc		;save loopcounter
	call &rand	;random value in a
	ld hl, BASE_SET
	add a, a	;16-bit adjustment
	ld c, a
	ld b, 0
	add hl, bc	;hl=pointer to element n
	ld a, e		;save loopcounter in a
	ex de, hl	;de=pointer to element n
	dec a		;N-1
	add a, a	;adjust for 16-bit
	ld c, a		;bc=2*(N-1)
	ld hl, POUCH+12
	add hl, bc	;hl = POUCH + 2*(N-1) + 12
	
	ld a, (de)	;copy element n from BaseSet to Pouch
	ld (hl), a
	inc de
	inc hl
	ld a, (de)
	ld (hl), a

	ld hl, BASE_SET+1
	add hl, bc	;hl points to BaseSet+1 + 2*(N-1) = last element in BaseSet
	ld a, (hl)	;copy last element to place n
	ld (de), a
	dec hl
	dec de
	ld a, (hl)
	ld (de), a

	pop bc		;get loopcounter
	djnz ScrambleLongLoop

ScramblingFinished:
	ld a, 1			;initialise phase number
	ld (PhaseNr), a
	ld (AnimCount1), a
	
;	ld (OldCPos), hl
	call &UnrollHistory
;	ld (CursorPos), hl
	ld hl, $0101
	call &DC_FirstTime	;initialize cursor functions
	call &MoveCursor	;start animation and draw cursor
	jp &DrawBoard

;************************************************************************************************
;*** ClearBoard: take all pieces from Board and reset score  
ClearBoard:
	ld hl, BOARD
	ld b, 96
ClearBoardLoop:
	ld (hl), 134	;set all board-values to 128 as marker for EMPTY_FIELD, set col to 6
	inc hl
	ld (hl), 0	;set shape to 0
	inc hl
	djnz ClearBoardLoop	;hl now points to next byte after Board (GRAPH_MEM+336 = 34705 = Score-2)

	ld b, 11		;clear 9 bytes: Score + Num4Ways + PiecesLeft + EndBonus + DispLegal
;	ld hl, Score
Init_ClearLoop:
	ld (hl), 0
	inc hl
	djnz Init_ClearLoop
	ld hl, POUCH
	ld (NextPiecePtr), hl	;store pointer to first piece in list
	ld a, 72
	ld (PiecesLeft), a
	ld hl, $0707		;store impossible piece at end of pouch
	ld (POUCH+144), hl	;to prevent display of legal-icons at end of game
	ret
;***********************************************************************************************
;*****  UnrollHistory: clear board and replay history of moves up to current one
UnrollHistory:
	call &ClearBoard
	ld hl, &EndGame
	ld (hl), $c9		;turn off endgame handler
	push hl			;save endgame-address

	xor a			;a=0
	ld hl, &DrawPiece
	ld (hl), $c9		;turn off drawing (ret)
	push hl			;save core-address
	ld hl, &HistOffs
	ld b, (hl)	;loopcounter
	ld (hl), a	;HistOffs=0, history is rewritten by PutPiece

	inc hl		;hl now points to start of History-table
UH_Loop:
	ld a, (hl)	;a=yx
	push hl		;save pointer to hist-table
	ld c, a
	and $f0
	rlca
	rlca
	rlca
	rlca		;a=y
	ld h, a		;h=y
	ld a, c
	and $0f		;a=x, carry=0, i.e. no checking
	ld l, a		;l=x
	push bc		;save loopcounter
	call &PutPiece	;place piece on board
	pop bc		;restore loopcounter
	pop hl		;restore index of hist-table
	inc hl		;next hist-entry
	djnz UH_Loop

	pop hl			;retrieve core-address
	ld (hl), 0		;set core=nop     i.e. turn on drawing

	pop hl			;retrieve endgame-address
	ld (hl), $cd		;turn on endgame-handler
	
	ld a, (PiecesLeft)
	cp 66
	jr nz, UH_NotFirst
	ld hl, $0101
	ld (CursorPos), hl	;cursor for first move in game is not floating over piece
UH_NotFirst:
;	call &DC_FirstTime
	

	ret

;***********************************************************************************************
;***************              BOARD MANAGEMENT ROUTINES                *************************
;***********************************************************************************************



;LegalScan - check if the current newpiece can be placed on field yx=hl
;store result in bit 4-6 of col of empty fields
;uses af, hl, de, bc
;output: (NBcount) is order of placement (0=No match  ...  4=FourWay)
LegalScan:
	xor a
	ld (MatchFlags), a	;clear MatchFlags
	ld (NBcount), a		;reset NBcount

	call &CalcBoardAddr	;hl=BoardAddr of field to be checked, de=yx
	ld (NextFieldPtr), hl	;save BoardAddr
	bit 7, (hl)		;field empty?
	ret z
	push de			;save yx
	inc hl
	push hl			;save BoardAddr+1
	;jr z, LS_NoMatch	;return if occupied: no placement possible
	ex de, hl		;hl=yx, de=BoardAddr
	
	ld b, a			;b=0 - counter for number of neigbourfields to identify CenterFields
TestTop:
	cp h		;y!=0 ?
	jr z, TestLeft
	inc b		;register neighbourfield regardless of state of occupation
	dec h		;test neighbour( x, y-1)
	call &CheckNeighbour
	inc h

TestLeft:
	cp l		;x!=0 ?
	jr z, TestBottom
	inc b
	dec l
	call &CheckNeighbour	;test neighbour (x-1, y)
	inc l

TestBottom:
	ld a, h		;a=y
	cp 7		;y!=7 ?
	jr z, TestRight
	inc b
	inc h
	call &CheckNeighbour	;test neighbour (x, y+1)
	dec h

TestRight:
	ld a, l		;a=x
	cp 11		;x!=11 ?
	jr z, LS_TestEnd
	inc b
	inc l
	call &CheckNeighbour	;test neighbour (x+1, y)
	;dec l
	
LS_TestEnd:
	ld a, b
	cp 4		;is this a centerfield ?
	jr nz, LS_EdgeField
	pop hl		;retrieve BoardAddr+1 into hl
	push hl		
	ld (hl), 1	;shape=1 -> CenterField

LS_EdgeField:
	ld a, (NBcount)
	or a	;NBcount=0 ?
	jr z, LS_NoMatch
	
	ld c, a		
	dec c		;adjust NBcount for use as array-index
	ld b, 0		;bc = nbcount
	ld hl, &AND_Table
	add hl, bc		;point to nb-offset
;	ld d, a			;save NBcount in d
	ld c, (hl)		;get offset
	add hl, bc		;add offset -> hl now points to start of table

	ld b, (hl)		;retrieve number of and-values to test
				;hl+1 points to body of table
LS_TestLoop:
	ld a, (MatchFlags)
	inc hl
	and (hl)	;mask out extra bits = normalize number of matches
	cp (hl)		;now compare with mask
	jr z, LS_Match
	djnz LS_TestLoop

LS_NoMatch:
	xor a
	ld (NBcount), a
	ld e, $ff		;erase NBcount -> legality flags = $ff
				;e=$ff -> offset for "no legality icon"
	jr LS_StoreInf
	
LS_Match:
	ld hl, &DispLegal
	ld a, (hl)	;get DispLegal-flag
	inc hl
	inc (hl)	;increment PossCount
	ld e, a
LS_StoreInf:
	pop hl		;retrieve BoardAddr+1=pointer to shape
	ld b, 6		;b=colour
	ld c, (hl)	;c=shape (=0: EgdeField or 1: CenterField)

	xor a
	or e		;e!=0? then draw legalIcon
	jr nz, LS_NoLegalIcon
	ld c, 2		;draw legal-icon (i.e. shape=2)


LS_NoLegalIcon:
	pop de		;retrieve yx
	call &DrawPiece

	ret		;match: (NBcount) = value of placement
	



;CheckNeighbour - check if piece stored at yx=hl matches with the NewPiece
;input:  (NextPiecePtr), initialized (NBcount)
;uses: de
;preserves: af, hl, bc
CheckNeighbour:
	push af
	push bc
	push hl		;save yx
	call &CalcBoardAddr	;hl=BoardAddr
	ld d, (hl)	;get colour
	bit 7, d	;Field occupied?
	jr nz, CN_End	;no, return
	ld a, d
	and 7		;mask out all bits that don't belong to the colour

	inc hl
	ld e, (hl)	;get shape
	ld hl, NBcount 
	inc (hl)	;register neighbour-piece
	ld b, (hl)	;b=NBcount

	ld hl, (NextPiecePtr)	;hl=pointer to Next Piece
	ld c, 0		;default: no match
	cp (hl)		;colour match?
	inc hl
	jr nz, NoColMatch
	
	ld c, 16	;set bit 4 for colour match
NoColMatch:
	ld a, (hl)	;a=NextPieceShape
	cp e		;shape match ?
	ld a, c
	jr nz, NoShapeMatch
	or 1		;set bit 0 for shape match
NoShapeMatch:
	or a		;any matches found?
	jr z, CN_End

	rrca		;shift one right because we can't use WHILE loops
	;inc b
ShiftLoop:		;shift left according to number of this neighbour
	rlca
	djnz ShiftLoop
	ld hl, MatchFlags
	or (hl)
	ld (hl), a
	
CN_End:
	pop hl		;restore yx
	pop bc
	pop af		;restore accu
	ret

;***************************************************************************************
PP_Quit:
;	pop hl		;restore stack	(BoardAddr)
	pop hl		;restore stack  (yx)
	ret

;***************************************************************************************
;Put current piece (pointed at by (NextPiecePtr)) at position hl=yx
;set carry flag to turn on checking
PutPiece:
	ld (CursorPos), hl	;set cursor to this placement (for UnrollHistory)
	push hl		;save yx
	push af
	ld de, &DrawPiece
	ld a, (de)		;get old drawing status
	push af
	ld a, $c9
	ld (de), a		;turn off drawing
	push de
	call &LegalScan		;(NBcount) now contains value of placement
	pop de
	pop af
	ld (de), a	;restore old drawing status
	pop af		;retrieve carry
	ld a, (NBcount)
	jr nc, PP_NoCheck
	or a		;test legality-marker
	jr z, PP_Quit
PP_NoCheck:

	ld hl, (NextFieldPtr)	;get boardaddr
	push hl			;and save it on stack

	or a
	jr z, PP_NoScore
			;now a contains nbcount
	ld c, a		
	sla c		;save NBcount*2 in c
	
	ld b, a		;NBcount=number of shifts

	inc hl		;hl points to shape
	ld a, 1
	cp (hl)		;shape=1->CenterField
	jr nz, PP_NoScore	;don't score edgefields
			;carry is now zero

	rra		;rotate low bit to carry
PP_ShiftLoop:
	rla
	djnz PP_ShiftLoop	;normal score=2^NBcount
	
	ld d, 0
	ld e, a		;de=normal score=2^NBcount
	ld hl, (Score)
	add hl, de	;hl=Score+normal points
	cp 8		;is this a fourway?
	jr nz, No4Way
	
	push hl		;save Score


	ld hl, Num4Ways
	inc (hl)
	ld a, (hl)	;l=Num4Ways (d is still 0)
	ld b, a		;save num4ways in a
	dec a
	cp 11		;a<12: no carry
	jr c, PP_lt12
	ld a, 10
PP_lt12:
	add a, a
	ld e, a		;e=2*(Num4Ways mod 12)

	ld hl, &BonusTable
	add hl, de	;hl now points to fourwaybonus
	ld c, (hl)
	inc hl
	ld b, (hl)	;bc=fourwaybonus
	pop hl		;retrieve Score
	add hl, bc	;hl=Score+FourWayBonus

No4Way:
	ld (Score), hl	;save score
	

PP_NoScore:	
	ld hl, (NextPiecePtr)	;get pointer to current piece
	ld b, (hl)	;get colour
	inc hl
	ld c, (hl)	;get shape
	inc hl
	ld (NextPiecePtr), hl	;store pointer to next piece
	pop hl		;hl=BoardAddr
	ld (hl), b	;save colour of piece on board
	inc hl
	ld (hl), c	;save shape of piece on board
	
	pop de		;de=yx, bc is still col/shape

;now compress yx-values and store in history list
	push bc		;save col/shape
	ld a, d		;a=y
	add a, a	;2y
	add a, a	;4y
	add a, a	;8y
	add a, a	;16y
	add a, e	;a=16y + x

	ld hl, &HistOffs
	ld b, 0
	ld c, (hl)		;bc=HistOffset
	inc (hl)		;point to next element
	inc hl		;hl now points to start of History-Table
	add hl, bc
	ld (hl), a		;store compressed yx-info
	pop bc		;retrieve col/shape

	ld hl, PiecesLeft
	dec (hl)
;	jr z, EndGame
	ld a, (&DrawPiece)
	or a		;drawing activated?
	jr nz, PP_NoDraw	;don't draw if drawing core is turned off

	call &DrawPiece	;display piece on screen

DrawNextPiece:
	ld a, (PiecesLeft)
	or a
	jr z, PP_NoDrawNext

	ld hl, (NextPiecePtr)
	ld b, (hl)	;get colour of next piece
	inc hl
	ld c, (hl)	;get shape
	ld de, $10d	;draw nextpiece at y=3, x=14
	call &DrawPiece

	ld hl, VIDEO_MEM+444		;erase background of numeric values
	ld a, 3
EraseValuesLoop1:
	ld bc, $0500	;b=5, c=0
EraseValuesLoop:
	ld (hl), c
	inc hl
	ld (hl), c
	inc hl
	ld (hl), c
	ld de, 14
	add hl, de
	djnz EraseValuesLoop
	ld e, 160
	add hl, de
	dec a
	jr nz, EraseValuesLoop1



PP_NoDrawNext:
	ld bc, 6		;3 values to display, loopc -= 2
	
UD_Loop:	
	ld hl, &DisplayTable-2	;table of cursor-locations
	add hl, bc
	ld e, (hl)		;get x-position
	inc hl
	ld d, (hl)		;get y-position
	ld (CURSOR_X), de	;save cursor-position
	
	ld hl, Score-2
	add hl, bc
	ld e, (hl)		;get lowbyte
	inc hl
	ld d, (hl)		;get highbyte
	ex de, hl		;hl=value to be displayed
	
	call DM_HL_DECI		;display value
	dec c
	dec c
	jr nz, UD_Loop

PP_LegalScan:
	xor a
	ld hl, &PossCount
	ld (hl), a	;reset counter of possible moves
	push hl
	ld ix, &LegalScan
	call &IterateFields	;check all fields for legality of next piece
	pop hl
	ld a, (hl)
	or a			;any possible placements found?
	ret nz			;if so, go back
				;otherwise, game is over
PP_NoDraw:
EndGame:
	;add handler for end of game here
	call CLEARLCD

	ld a, (PiecesLeft)
	cp 3
	jr nc, NoEndBonus
	add a, a
	ld c, a
	ld b, 0
	ld hl, &EndBonusTable
	add hl, bc
	ld c, (hl)
	inc hl
	ld b, (hl)
	ld hl, (Score)
	add hl, bc
	ld (Score), hl
	ld (EndBonus), bc
	
	
NoEndBonus:
	ld a, (&Modus)
	or a			;no highscore check in practice mode
	jr z, EG_Print
	ld hl, (&HighScore)
	ld de, (Score)
	call CP_HL_DE
	jr nc, EG_Print		;new highscore?

	ld (&HighScore), de	;save new highscore
	ld hl, (Num4Ways)
	ld (&High4Way), hl	;save new best four-way-value


	ld b, 6
	ld hl, &GameOverTable0	;display new highscore-report
	jr PrintStringTable2


EG_Print:
	ld b, 5		;print 5 strings
	ld hl, &GameOverTable
PrintStringTable2:
	ld e, (hl)		;get y-position
	inc hl
	ld d, (hl)		;get x-position
	inc hl
	ld (CURSOR_ROW), de
	call D_ZT_STR
	djnz PrintStringTable2	;hl now points at next byte after GameOverTable, i.e. the cursor-positions

	push hl		;save cursor positions
	ld hl, Score
	ld bc, $0f00	;b=x=15  c=loopcounter
EG_DispLoop:
	ex (sp), hl	;hl=pointer to next cursor position, (sp)=pointer to next value
	ld c, (hl)	;get y-position
	inc hl
	ex (sp), hl	;hl=pointer to next value, (sp)=pointer to next cursor position

	ld (CURSOR_ROW), bc
	ld e, (hl)
	inc hl
	ld d, (hl)
	inc hl
	push hl
	ex de, hl
	call D_HL_DECI
	pop hl
	ld a, c
	cp 3		;test for last value in cursorpos-table
	jr nz, EG_DispLoop
	pop hl		;restore sp

	call &Wait
	
	ld a, $c9
	ld (&EndGame), a	;don't enter EndGame again until next game is started

;	xor a
;	ld (&AnimCore), a	;stop cursor
	
	pop hl			;restore sp and don't return from call
	jr DrawBoard

	
;***************************************************************************************
;**** Print StringTable: Print in menu style the table of zero-terminated strings with 
;**** leading cursor-information
;**** hl=start of stringtable      b=number of strings
PrintStringTable:
	ld e, (hl)		;get x-position
	inc hl
	ld d, (hl)		;get y-position
	inc hl
	ld (CURSOR_X), de
	;ld (CURSOR_ROW), de
PST_Core:			;core+1:  $4a=menu string    $5f=normal string
	call D_ZM_STR
	djnz PrintStringTable
	ret


;***************************************************************************************
;Draws the pouch

DP_Empty:
	ld hl, &PouchEmptyStr
	ld b, 1
	call &PrintStringTable
	jr DP_End

DrawPouch:
	call CLEARLCD

	ld hl, (NextPiecePtr)
	ld de, $0102	;x=2, y=1
	ld b, 72

	

DrawPouchLoop:
	ld a, (PiecesLeft)
	or a		;pouch empty?
	jr z, DP_Empty
	cp b
	jr c, DP_NextPiece

	push bc		;save loopcounter
	ld b, (hl)	;get colour
	inc hl
	ld c, (hl)	;get shape
	inc hl
	
	
	push hl
	push de
	call &DrawPiece	;draw piece(col=h, shape=l) on screen(y=d, x=e)
	pop de
	pop hl
	pop bc

DP_NextPiece:
	inc e	;next x-position
	ld a, 14
	cp e
	jr nz, DP_NoCRLF	;end of line reached?
	ld e, 2		;reset x = cr
	inc d		;next y	= nl
DP_NoCRLF:
	djnz DrawPouchLoop
DP_End:
	call &Wait
;*****************************************************************************************************
DrawBoard:
;	ld c, 16
;	ld hl, VIDEO_MEM
;FillLoop1:
;	ld b, 64
;FillLoop0:
;	ld (hl), 255
;	inc hl
;	djnz FillLoop0
;	dec c
;	jr nz, FillLoop1


	call CLEARLCD
	ld de, &AboutBmp
	ld hl, VIDEO_MEM+15
	ld b, 64
	call &GD_Loop		;draw about-bitmap at right edge of screen


	ld ix, &DrawBoardField
	call &IterateFields

;	call &Wait
	
	ld b, 4			;display 4 strings
	ld hl, &StringTable
	call &PrintStringTable

	call &DrawNextPiece
	jp &KeyLoop
;******************************************************************
;go through all boardfields and call function (ix)
;changes hl
IterateFields:
	
	ld h, 8		;8 vertical fields
IterBoardLoopX:
	ld l, 12	;12 horizontal fields
IterBoardLoopY:
	push hl		;save loopcounter
	dec h		;adjust counter
	dec l
	call &IterateJump		;save PC and go to function (IterateCore)
	pop hl		;retrieve loopcounter
	dec l
	jr nz, IterBoardLoopY	;next y
	dec h
	jr nz, IterBoardLoopX	;next x\
	ret

IterateJump:
	jp (ix)
;**********************************BOARD IS NOW DRAWN


KL_MainMenu:
	call ClearLCD
	ld hl, 0
	ld (CURSOR_ROW), hl	;cursorpos=0
	ld hl, &HelpScreenTable
	set 3, (IY+05)		;display inverse text
	call D_ZT_STR		;print (hl)
	res 3, (IY+05)		;display normal text
	

	ld b, 15
	call &PrintStringTable	;last entry in table should set cursor to x=0, y=48

	ld hl, (&HighScore)	;print highscore with associated number of fourways
	call DM_HL_DECI
	ld hl, CURSOR_X
	ld (hl), 44		;set x-position to 48
	ld hl, (&High4Way)
	call DM_HL_DECI

	call &Wait		;now a contains scancode of pressed key
	ld de, &HistOffs

	cp K_F3
	jr nz, MM_NotF3

	;now restart current game
	ex de, hl
	ld (hl), 6	;reset history
	jp &ScramblingFinished


MM_NotF3:
	ld hl, &Modus		;prepare to change modus-variable

	cp K_F1			;start New Tournament Game?
	jr nz, MM_NotF1
	ld (hl), $ff		;set modus to Tournament
	inc hl
	ld (hl), $ff		;turn off display of legal icons
	jr KL_NewGame

MM_NotF1:
	cp K_F2			;start new practice game?
	jr nz, MM_NotF2
	ld (hl), 0		;set modus to practice
	inc hl
	ld (hl), 0		;turn on display of legal icons

KL_NewGame:
	ex de, hl	;hl=HistOffs
	ld (hl), 6

	call RANDOM	;get random number as seed
	ld hl, seed1
	ld de, &oldSeeds
	ld (hl), a	;set random number generator
	ldi
	ldi
	ldi
	ldi		;overwrite oldSeeds
	jp &Init_Pouch
	



MM_NotF2:
	jp &DrawBoard

KL_GoBack:
	ld hl, &HistOffs
	ld a, 6
	cp (hl)		;histoffs must be greater than 6 to allow going back
	jr nc, KeyLoop
	dec (hl)
KL_CallUnrHist:
	call &UnrollHistory
	call &DrawNextPiece
;	jr DrawBoard
	jr CursorHasMoved2


MoveDown:
	ld a, 7
	cp h		;y<7 ?
	jr z, KeyLoop	;y==11, do not go down any further
	inc h
	jr CursorHasMoved
MoveUp:
	xor a		;a=0
	cp h		;y>0 ?
	jr z, KeyLoop	;y==0, do not go up any further
	dec h
	jr CursorHasMoved

MoveLeft:
	xor a		;a=0
	cp l		;x>0?
	jr z, KeyLoop	;x==0, don't go further left
	dec l
	jr CursorHasMoved

MoveRight:
	ld a, 11
	cp l
	jr z, KeyLoop	;x==11, don't go further right
	inc l
	;proceed to CursorHasMoved

CursorHasMoved:
	ld (CursorPos), hl
CursorHasMoved2:
	call &MoveCursor
	jr KL_Anim
KeyLoop:
	call GET_KEY		
			;now a contains scancode of a key
	cp K_EXIT
	ret z
	ld b, a		;save a in b
	
	ld a, (&Modus)
	or a		;test modus
	ld a, b		;restore a
	jr nz, KL_Tournament

	cp K_DEL
	jr z, KL_GoBack
	cp K_ALPHA
	jp z, &DrawPouch
	cp K_F1
	jr z, KL_SwitchLegal

KL_Tournament:
	ld a, (&EndGame)		;is the game still going?
	cp $c9
	ld a, b		;restore a
	jr nz, KL_NotGameFinished
	or a				;if not, any key jumps to Main Menu
	jp nz, &KL_MainMenu
	jr KeyLoop

KL_NotGameFinished:
;	cp K_CLEAR
;	call z, &EndGame
	ld hl, (CursorPos)	;h=y, l=x
	cp K_UP
	jr z, MoveUp
	cp K_DOWN
	jr z, MoveDown
	cp K_LEFT
	jr z, MoveLeft
	cp K_RIGHT
	jr z, MoveRight
	cp K_SECOND
	jr z, KL_Second
	cp K_ENTER
	jr z, KL_Second
	cp K_F2
	jr z, KL_StopCursor
	cp K_MORE
	jp z, &KL_MainMenu

KL_AnimStart:
	ex de, hl		;save cursor position
	ld hl, AnimCount	;animation delay counter
AnimCore:
	dec (hl)
	jr nz, KeyLoop
	inc hl			;hl=AnimCount1
	dec (hl)
	jr nz, KeyLoop
	ld (hl), 8		;wait for 8*256 keyloops next time
	ex de, hl		;retrieve cursor position
KL_Anim:
	call &AnimCursor
	jr KeyLoop	;rotate cursor

KL_StopCursor:
	ld hl, &AnimCore
	ld a, $35
	xor (hl)
	ld (hl), a	;set flag to stop cursor rotation
	jr KeyLoop

KL_Second:
	scf		;turn on checking if piece is legal
;   ccf		;turn off checking
	call &PutPiece
	jr KeyLoop

KL_SwitchLegal:
	ld hl, &DispLegal
	ld a, (hl)
	cpl		;switch legal-status
	ld (hl), a
	call &PP_LegalScan	;iterate fields with legalscan
	jr KL_Anim


Wait:	call GET_KEY
	or a		;test for any key
	jr z, Wait
	ret


;*************************************************************************************************************************
;Copies the cursor to location stored in (CursorPos)
;call MoveCursor if the cursorposition has changed
MoveCursor:
	ld hl, &AnimCore
	ld (hl), $35		;start cursor rotation

	ld c, 0			;draw white frame
	call &RightLine

DC_FirstTime:
	ld de, (CursorPos)	;retrieve current cursor position
	ld hl, OldCPosY
	ld (hl), d		;store old y-value

	call &CalcVideoAddr	;hl now contains video-addr(y, x), de=yx
	ld bc, -16
	add hl, bc
	ld (OldCPos), hl	;store this for restoration

AnimCursor:
	ld c, 128	;draw black frame
	call &RightLine
		
	ld hl, PhaseNr
	dec (hl)
	jr nz, DC_IncPhase
	ld (hl), 4	;set phasenr to 4
	ld hl, &Cursor	;equiv. to phase0
	ld (PhasePtr), hl
	jr DC_Draw

DC_IncPhase:
	ld hl, (PhasePtr)
	ld de, 8
	add hl, de
	ld (PhasePtr), hl
	
DC_Draw:
	;ld hl, (PhasePtr)
	;ld hl, &Cursor
	ld b, 6			;get loopcount
	ld c, (hl)		;get first byte, whether or not it is used
	inc hl
	
	ld a, (OldCPosY)	;get y-position
	or a
;	cp 1
	jr z, DC_InFirstLine	;y=0?
	ex de, hl
	ld hl, (OldCPos)	;hl=screen-address
	ld (hl), c		;write first cursor byte on screen
	ex de, hl		;restore pointer to phase-table
	jr DC_cont

DC_InFirstLine:
	dec b			;skip first byte
	inc hl			;in table

DC_cont:
	ld d, 0
	ld e, (hl)		;de=offset
	inc hl			;next phase-byte
	push hl
	ld hl, (OldCPos)	;restore starting Screen-Addr
	add hl, de		;add offset
	res 7, (hl)		;clear leftmost bit
	ex de, hl		;store calculated screen-address in de
	pop hl			;restore phase-ptr
	djnz DC_cont
	
	;inc hl
	ld a, (hl)		;get last phase-byte
	ld (de), a		;and copy on screen

	ret

;Draw lines around the field pointed at by hl
;c contains mask: 0=white line   128=black line
;output: af destroyed
;	 hl points to next byte below draw block
RightLine:
	ld hl, (OldCPos)	;old videoaddr of cursor
	ld de, 16

	ld a, (OldCPosY)
	or a
	jr z, RL_FirstLine	;y=0?


	ld (hl), c
	inc hl
	set 7, (hl)	;default: black
	xor a
	or c		;d=0 ?
	jr nz, RL_NoBlack0
	res 7, (hl)	;exception: white
RL_NoBlack0:
	dec hl
	
RL_FirstLine:
	add hl, de
	dec e		;de=15
	
	ld b, 7
RL_InnerLoop:
	ld a, (hl)
	and 127		;clear leftmost bit
	or c		;and set if desired
	ld (hl), a
	inc hl		;next byte
	ld a, (hl)
	and 127		;clear leftmost bit
	or c		;and set if desired
	ld (hl), a

	add hl, de	;hl+=15
	djnz RL_InnerLoop

	ld (hl), c
	inc hl
	res 7, (hl)
	xor a
	or c		;d=0?
	jr z, RL_White
	set 7, (hl)
RL_White:
	ret


	
;*******************************************************************************************
;Draw a board field with hl=yx2+0
;uses af, hl, bc, de
DrawBoardField:
	call &CalcBoardAddr	;hl=BoardAddr

	ld b, (hl)	;get colour
	inc hl
	ld c, (hl)	;get shape
			;proceed to DrawPiece

;*******************************************************************************************
;Draws the piece col/shape = bc
;at yx = de 
;destroys af, hl, bc, de
DrawPiece:
	nop		;nop=0 vs. ret = $c9
	push de		;save yx ?
	ld hl, &Pieces-42
	ld de, 42	;one row of pieces = 6*7 byte
	ld a, b		;mask out upper bits of colour byte
	and 7
	ld b, a
	inc b
CountCol:
	add hl, de
	djnz CountCol
			;b is now zero
	ld a, c
	add a, a	;shape*2
	add a, a	;shape*4
	add a, a	;shape*8
	sub c		;shape*7
	ld c, a		;bc=shape*7	since b=0
	add hl, bc	;hl = bitmap-addr

	pop de		;b=y, c=x
	push hl		;save bitmap-addr
	call &CalcVideoAddr	;hl=video-addr
	pop de		;de=bitmap-addr

;Drawing Routine: b=number of bytes, hl=VideoAddr, de=bitmapAddr
GeneralDraw:
	ld b, 7		;loopcounter
GD_Loop:
	ld a, (de)	;get bitmap-byte
	inc de
	ld (hl), a	;copy on screen
	push de		;hl += 16
	ld de, 16
	add hl, de
	pop de
	djnz GD_Loop

	ret

;**********************************************************************************************************
;input: d=y e=x
:output: hl=VideoAddr, de still yx
;destroys af and bc
CalcVideoAddr:
	ld a, d		;a=y<<8
	srl a		;a=y<<7 (LSB in carry)
	ld h, a		;h=y<<7
	ld a, 0		;a=0, cannot use xor a because carry must be preserved
	rra		;bit 7 = carry
	add a, e	;add x	 (x <= 11, hence no carry)
	ld l, a		;hl=128y + x
	ld bc, VIDEO_MEM
	add hl, bc	;hl = VideoAddr

	ret
;**********************************************************************************************************
;input: hl = yx
;output: hl=pointer to BOARD+2(12y+x)
;	 de=yx
;destroys bc
CalcBoardAddr:
	push hl
	ld b, h
	ld c, l		;bc=yx
	ld hl, BOARD-24
	ld de, 24
	inc b
PP_CountYLoop:
	add hl, de
	djnz PP_CountYLoop	;b=0, hl=BOARD+24*y

	sla c		;b=0, hence bc=2*x
	add hl, bc	;hl=BOARD+24y+2x = BOARD+2(12y+x) = BoardAddr
	pop de
	ret

;*****************************************************************************************
;Scramble list in (hl) with 6 elements and copy to (de), double spaced
;Addr of source list in hl
;Addr of destination list in de
;uses af, hl, bc, de
ScrambleShortLst:	
	ld c, 6			;init loopcounter
	ld (SrcAddr), hl
	ld (DstAddr), de
ScrambleShortLoop:
	ld e, c			;upper limit for rand
	call &rand		;a now contains random number with 0 <= a < e

	ld b, 0			
	ld c, a			;get random number in c
	ld hl, (SrcAddr)
	push hl			;save SrcAddr
	add hl, bc		;hl now points to SrcLst + rand
	ld a, e			;save loopcounter in a
	ex de, hl		;save hl in de
	
	ld hl, (DstAddr)
	ld c, a			;copy loopcounter to c
	sla c			;multiply by two for double-spaced list
	add hl, bc
				;hl now contains DstLst + 2*loopcounter
				;de contains SrcLst + rand
	ld c, a			;save loopcounter in c

	ld a, (de)		;copy randomly chosen element to DstLst
	ld (hl), a	
	
	pop hl			;retrieve SrcAddr	
	dec c			;decrement loopcounter and set zero-flag
	add hl, bc		;add loopcounter-1 -> hl now points to last element in list
	ld a, (hl)
	ld (de), a		;copy last element of list to the now empty position
	
	jr nz, ScrambleShortLoop

	ret
;****************************************************************************************************

rand:	;returns random number between 0 and e in a
	;uses af, hl, bc
	;de is unchanged
	call &random1	;get 8-bit random number in a
	ld h, a
	ld l, e
;proceed to division

	
	
;unsigned 8-bit division
;divide h through l
;remainder in a
;result in h
;uses a, hl, bc

UDiv8:
	ld b, 8		;8 bits in dividend
	xor a		;clear carry and remainder

DivLoop:
;rotate the carry into c and all following registers
 	rl h		;rotate left dividend
	rla		;rotate left remainder

 	ld c, a		;save a
 	sbc a, l	;subtract divisor and carry from remainder
 	ccf		;complement carry 

 	jr c, Drop
 	ld a, c		;restore a

Drop:
  	djnz  DivLoop

 	rl h
;h now contains the result
;a contains the remainder
 	ret



;thanks to Risto Jaervinen for the random number generator
;the original source of this is in the file rnds.asm from ticalc.org
;
;       High quality 2x16-bit seed random number generator
;       --------------------------------------------------
;   May be slow but sure is 'random'.
;   Routine uses two 16-bit variables, seed1 and seed2.
;   Read and cut the result out of seed1.
; uses: af, bc, hl
; returns random value in a

random1:
   	ld   hl,(seed2)              ;Load seed2 to HL
   	ld   a,h
   	add  a,a                     ;Set highest bit of seed to carry
   	rl   l                       ;rotate L left (C<=L<=C)
   	rl   h                       ;rotate H left (C<=L<=C)
   	add  a,a                     ;Set second highest bit of seed to carry
   	rl   l
   	rl   h
   	add  a,a
   	rl   l
   	rl   h
   	ld   bc,$7415
   	add  hl,bc                   ;Add $7415 to HL
   	ld   (seed2),hl              ;Load seed2 with HL

   	ld   a,l                     ;Preserve L
   	ld   hl,(seed1)              ;Load seed1 to HL
   	and  %00001111
   	jr   z,norot
   	ld   b,a
rotmore:                        ;Rotate HL (a and %1111) bits
   	add  hl,hl                   ;Z80-version of 16-bit rotate!
   	jr   nc,notinc
   	set  0,l
notinc:
   	djnz rotmore
norot:
   	ld   bc,$6349                ;Add $6349 to HL
   	add  hl,bc
   	ld   (seed1),hl              ;Load seed1 with HL

	ld a, h		;return random value in a
   	ret

;seed1:	.db 12, 34
;seed2:	.db 56, 78

oldSeeds: .db 12, 34, 56, 78

Bitmap:
Pieces:
white:
.db $7F, $41, $41, $41, $41, $41, $7F
.db $1C, $22, $41, $41, $41, $22, $1C
.db $08, $14, $22, $41, $22, $14, $08
.db $00, $7F, $41, $41, $41, $7F, $00
.db $3E, $22, $22, $22, $22, $22, $3E
.db $08, $14, $22, $41, $22, $41, $7F

whcross:
.db $7F, $77, $77, $41, $77, $77, $7F
.db $1C, $36, $77, $41, $77, $36, $1C
.db $08, $14, $36, $41, $36, $14, $08
.db $00, $7F, $77, $41, $77, $7F, $00
.db $3E, $36, $36, $22, $36, $36, $3E
.db $08, $14, $36, $41, $36, $77, $7F

hline:
.db $7F, $41, $41, $7F, $41, $41, $7F
.db $1C, $22, $41, $7F, $41, $22, $1C
.db $08, $14, $22, $7F, $22, $14, $08
.db $00, $7F, $41, $7F, $41, $7F, $00
.db $3E, $22, $22, $3E, $22, $22, $3E
.db $08, $14, $22, $7F, $22, $41, $7F

vline:
.db $7F, $49, $49, $49, $49, $49, $7F
.db $1C, $2A, $49, $49, $49, $2A, $1C
.db $08, $1C, $2A, $49, $2A, $1C, $08
.db $00, $7F, $49, $49, $49, $7F, $00
.db $3E, $2A, $2A, $2A, $2A, $2A, $3E
.db $08, $1C, $2A, $49, $2A, $49, $7F

grey:
.db $7F, $55, $6B, $55, $6B, $55, $7F
.db $1C, $2A, $55, $6B, $55, $2A, $1C
.db $08, $14, $2A, $55, $2A, $14, $08
.db $00, $7F, $55, $6B, $55, $7F, $00
.db $3E, $2A, $36, $2A, $36, $2A, $3E
.db $08, $14, $2A, $55, $2A, $55, $7F

black:
.db $7F, $7F, $7F, $7F, $7F, $7F, $7F
.db $1C, $3E, $7F, $7F, $7F, $3E, $1C
.db $08, $1C, $3E, $7F, $3E, $1C, $08
.db $00, $7F, $7F, $7F, $7F, $7F, $00
.db $3E, $3E, $3E, $3E, $3E, $3E, $3E
.db $08, $1C, $3E, $7F, $3E, $7F, $7F

EdgeField:
.db $00, $2A, $00, $2A, $00, $2A, $00		;col=6    shape=0

CenterField:
.db 0, 0, $14, 0, $14, 0, 0			;col=6    shape=1

LegalIcon:
.db %01100011, %01000001, 0, 0, 0, %01000001, %01100011	;col=6 shape=2


Cursor:
;.db $DD, 32, 01, 63, 01, 31, $DD
;.db $EE, 16, 33, 31, 33, 15, $BB
;.db $77, 01, 63, 01, 64,  0, $77
;.db $BB, 17, 31, 33, 31, 16, $EE

.db $DD, 32, 32, 33, 96,  97, 128, $DD
.db $EE, 16, 16, 49, 80, 113, 128, $BB
.db $77,  1, 64, 64, 65, 129, 128, $77
.db $BB, 17, 17, 48, 81, 112, 128, $EE

AboutBmp:
.db $4A, $55, $4F, $40, $4F, $49, $46, $40
.db $4F, $49, $46, $40, $5F, $48, $44, $48
.db $5F, $40, $40, $46, $49, $46, $40, $5F
.db $49, $46, $40, $46, $49, $46, $40, $4A
.db $55, $5F, $40, $40, $4C, $42, $4D, $40
.db $46, $49, $5F, $40, $40, $5F, $51, $5F
.db $40, $4E, $51, $5F, $40, $5F, $40, $5F
.db $44, $5F, $40, $52, $55, $49, $40, $5F

AND_Table:
.db 4, 6, 8, 14
.db 2, %00010000, %00000001		;one neighbour: colour or shape must match
.db 2, %00010010, %00100001		;two neighbours: match col/shape or shape/col
.db 6, %00010110, %00100101, %01000011,  %00110100, %01010010, %01100001
.db 6, %00111100, %01011010, %10010110,  %01101001, %10100101, %11000011

BonusTable:
.dw 25, 50, 100, 200, 400, 600, 800, 1000
.dw 2000, 5000, 10000 

EndBonusTable:
.dw 1000, 500, 100

StringTable:
.db 102, 0, "Next", 0
.db 104, 20, "Left", 0
.db 100, 35, "Score", 0
.db 100, 50, "4Ways", 0

DisplayTable:
.db 100, 41	;Score
.db 94, 56	;4Ways
.db 96, 26	;Left

Modus:
.db 0	;switch for Tournament/Practice mode
DispLegal:
.db 0
PossCount:
.db 0

HistOffs:
.db 6		;offset-pointer to current entry in history table

History:				;reserve 4*9*2=72 byte for history
.db 0, 123, 112, 11, 53, 70		;compressed positions of first six pieces
.dw 0, 0, 0, 0, 0, 0
.dw 0, 0, 0, 0, 0, 0, 0, 0, 0
.dw 0, 0, 0, 0, 0, 0, 0, 0, 0
.dw 0, 0, 0, 0, 0, 0, 0, 0, 0

PouchEmptyStr:
.db 41, 29, "Pouch is empty", 0


HelpScreenTable:
.db "      ISHIDO 1.0     ", 0
.db 15, 58, "by Bodo.Maass@privat.post.de", 0
.db  0,  8, "--New Game--", 0
.db  0, 14, "F1: Tournament", 0
.db  0, 20, "F2: Practice", 0
.db  0, 26, "F3: Restart", 0
.db 0, 42, "HighScore:", 0
.db 22, 48, "Points", 0
.db 66, 48, "FourWays", 0
.db 61,  8, "--Game Controls--", 0
.db 58, 14, "F1: Show poss. moves", 0
.db 58, 20, "F2: Stop cursor", 0
.db 58, 26, "2nd/ENTER: Put Piece", 0
.db 58, 32, "ALPHA: Show pouch", 0
.db 58, 38, "DEL: Undo moves", 0
.db 1, 48, 0

HighScore:
.dw 0		;highscore
High4Way:
.dw 0

GameOverTable0:
.db 7, 1, "** NEW HIGHSCORE **", 0

GameOverTable:
.db 0, 6, "Game Over!", 0
.db 2, 1, "Pieces left:", 0
.db 3, 1, "Endbonus:", 0
.db 4, 1, "FourWays:", 0
.db 5, 1, "Total Score:", 0
.db 5, 4, 2, 3		;cursor-positions for gameover-display

.end




